#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
Module to produce the wrapper function to be executed by the executors.
"""

import builtins as __builtin__
import inspect
import json
import traceback

from hops import hdfs as hopshdfs
from hops.experiment_impl.util import experiment_utils

from maggy import util, tensorboard
from maggy.core import rpc, exceptions
from maggy.core.reporter import Reporter


def _prepare_func(
    app_id,
    run_id,
    experiment_type,
    train_fn,
    server_addr,
    hb_interval,
    secret,
    optimization_key,
    log_dir,
):
    def _wrapper_fun(iter):
        """
        Wraps the user supplied training function in order to be passed to the
        Spark Executors.

        Args:
            iter:

        Returns:

        """
        experiment_utils._set_ml_id(app_id, run_id)

        # get task context information to determine executor identifier
        partition_id, task_attempt = util.get_partition_attempt_id()

        client = rpc.Client(
            server_addr, partition_id, task_attempt, hb_interval, secret
        )
        log_file = (
            log_dir
            + "/executor_"
            + str(partition_id)
            + "_"
            + str(task_attempt)
            + ".log"
        )

        # save the builtin print
        original_print = __builtin__.print

        reporter = Reporter(log_file, partition_id, task_attempt, original_print)

        def maggy_print(*args, **kwargs):
            """Maggy custom print() function."""
            original_print(*args, **kwargs)
            reporter.log(" ".join(str(x) for x in args), True)

        # override the builtin print
        __builtin__.print = maggy_print

        try:
            client_addr = client.client_addr

            host_port = client_addr[0] + ":" + str(client_addr[1])

            exec_spec = {}
            exec_spec["partition_id"] = partition_id
            exec_spec["task_attempt"] = task_attempt
            exec_spec["host_port"] = host_port
            exec_spec["trial_id"] = None

            reporter.log("Registering with experiment driver", False)
            client.register(exec_spec)

            client.start_heartbeat(reporter)

            # blocking
            trial_id, parameters = client.get_suggestion(reporter)

            while not client.done:
                if experiment_type == "ablation":
                    ablation_params = {
                        "ablated_feature": parameters.get("ablated_feature", "None"),
                        "ablated_layer": parameters.get("ablated_layer", "None"),
                    }
                    parameters.pop("ablated_feature")
                    parameters.pop("ablated_layer")

                tb_logdir = log_dir + "/" + trial_id
                trial_log_file = tb_logdir + "/output.log"
                reporter.set_trial_id(trial_id)

                # If trial is repeated, delete trial directory, except log file
                if hopshdfs.exists(tb_logdir):
                    util._clean_dir(tb_logdir, [trial_log_file])
                else:
                    hopshdfs.mkdir(tb_logdir)

                reporter.init_logger(trial_log_file)
                tensorboard._register(tb_logdir)
                if experiment_type == "ablation":
                    hopshdfs.dump(
                        json.dumps(ablation_params, default=util.json_default_numpy),
                        tb_logdir + "/.hparams.json",
                    )

                else:
                    hopshdfs.dump(
                        json.dumps(parameters, default=util.json_default_numpy),
                        tb_logdir + "/.hparams.json",
                    )

                try:
                    reporter.log("Starting Trial: {}".format(trial_id), False)
                    reporter.log("Trial Configuration: {}".format(parameters), False)

                    if experiment_type == "optimization":
                        tensorboard._write_hparams(parameters, trial_id)

                    sig = inspect.signature(train_fn)
                    if sig.parameters.get("reporter", None):
                        retval = train_fn(**parameters, reporter=reporter)
                    else:
                        retval = train_fn(**parameters)

                    if experiment_type == "optimization":
                        tensorboard._write_session_end()

                    retval = util._handle_return_val(
                        retval, tb_logdir, optimization_key, trial_log_file
                    )

                except exceptions.EarlyStopException as e:
                    retval = e.metric
                    reporter.log("Early Stopped Trial.", False)

                reporter.log("Finished Trial: {}".format(trial_id), False)
                reporter.log("Final Metric: {}".format(retval), False)
                client.finalize_metric(retval, reporter)

                # blocking
                trial_id, parameters = client.get_suggestion(reporter)

        except:  # noqa: E722
            reporter.log(traceback.format_exc(), False)
            raise
        finally:
            reporter.close_logger()
            client.stop()
            client.close()

    return _wrapper_fun



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


class AblationStudy(object):
    """The `AblationStudy` object is the entry point to define an ablation
    study with maggy. This object can subsequently be passed as an argument
    when the experiment is launched with `experiment.lagom()`.

    Sample usage:

    >>> from maggy.ablation import AblationStudy
    >>> ablation_study = AblationStudy('titanic_train_dataset',
    >>>     label_name='survived')

    The above code will create an `AblationStudy` instance with a default
    dataset generator function, which uses the project feature store to
    return a `TFRecordDataset` based on the feature ablation configuration
    (for an example, look at `ablator.LOCO.get_dataset_generator()`).
    If you want to provide your own dataset generator function,
    define it before creating the `AblationStudy` instance and pass it to
    the initializer or use 'set_dataset_generator()'. In the example below
    we assume the user has created a function called `create_tf_dataset()`
    that returns a `TFRecordDataset`:

    >>> ablation_study = AblationStudy('titanic_train_dataset',
            label_name='survived', dataset_generator=create_tf_dataset)

    In case you want to perform feature ablation with your custom dataset
    generator function, then of course your function should be able to return
    specific datasets based on the feature ablation configuration.
    For an example implementation of such logic, look at
    `ablator.LOCO.get_dataset_generator()`.

    After creating your `AblationStudy` instance, you should define your study
    configuration by including layers and features that you want to be ablated:

    >>> ablation_study.features.include('pclass', 'fare')
    >>> ablation_study.model.layers.include('my_dense_two',
    >>>     'my_dense_three')

    You can also add a layer group using a list:

    >>> ablation_study.model.layers.include_groups(['my_dense_two',
    >>>     'my_dense_four'])

    Or add a layer group using a prefix:

    >>> ablation_study.model.layers.include_groups(prefix='my_dense')

    Next you should define a base model function using the layer and feature
    names you previously specified:

    >>> # you only need to add the `name` parameter to layer initializers
    >>> def base_model_generator():
    >>>     import tensorflow as tf
    >>>     model = tf.keras.Sequential()
    >>>     model.add(tf.keras.layers.Dense(64, activation='relu'))
    >>>     model.add(tf.keras.layers.Dense(..., name='my_dense_two', ...)
    >>>     model.add(tf.keras.layers.Dense(32, activation='relu'))
    >>>     model.add(tf.keras.layers.Dense(..., name='my_dense_sigmoid', ...)
    >>>     # output layer
    >>>     model.add(tf.keras.layers.Dense(1, activation='linear'))
    >>>     return model

    Make sure to include the generator function in the study:

    >>> ablation_study.model.set_base_model_generator(base_model_generator)

    Last but not least you can define your actual training function:

    >>> from maggy import experiment
    ​
    >>> def training_function(dataset_function, model_function):
    >>>     import tensorflow as tf
    >>>     epochs = 5
    >>>     batch_size = 10
    >>>     tf_dataset = dataset_function(epochs, batch_size)
    >>>     model = model_function()
    >>>     model.compile(optimizer=tf.train.AdamOptimizer(0.001),
    >>>             loss='binary_crossentropy',
    >>>             metrics=['accuracy'])

    >>>     history = model.fit(tf_dataset, epochs=epochs, steps_per_epoch=30)
    >>>     return float(history.history['acc'][-1])

    Lagom the experiment:

    >>> result = experiment.lagom(train_fn=training_function,
    >>>                         experiment_type='ablation',
    >>>                         ablation_study=ablation_study,
    >>>                         ablator='loco',
    >>>                         name='Titanic-LOCO')
    """

    def __init__(
        self, training_dataset_name, training_dataset_version, label_name, **kwargs
    ):
        """Initializes the ablation study.

        :param training_dataset_name: Name of the training dataset in the
            featurestore.
        :type training_dataset_name: str
        :param training_dataset_version: Version of the training dataset to be
            used.
        :type training_dataset_version: int
        :param label_name: Name of the target prediction label.
        :type label_name: str
        """
        self.features = Features()
        self.model = Model()
        self.hops_training_dataset_name = training_dataset_name
        self.hops_training_dataset_version = training_dataset_version
        self.label_name = label_name
        self.custom_dataset_generator = kwargs.get("dataset_generator", False)

    def to_dict(self):
        """
        Returns the ablation study configuration as a Python dictionary.

        :return: A dictionary with ablation study configuration parameters as
            keys (i.e. 'training_dataset_name', 'included_features', etc.)
        :rtype: dict
        """
        ablation_dict = {
            "training_dataset_name": self.hops_training_dataset_name,
            "training_dataset_version": self.hops_training_dataset_version,
            "label_name": self.label_name,
            "included_features": list(self.features.included_features),
            "included_layers": list(self.model.layers.included_layers),
            "custom_dataset_generator": True
            if self.custom_dataset_generator
            else False,
        }

        return ablation_dict

    def set_dataset_generator(self, dataset_generator):
        """
        Sets a user-defined function as the dataset generator.

        :param dataset_generator: a function that returns the dataset(s) required for the training loop.
        """
        self.custom_dataset_generator = dataset_generator


class Features(object):
    def __init__(self):
        self.included_features = set()

    def include(self, *args):
        """
        Add one or several features of the dataset to the ablation study. Features can be included one by one or
        as a list of strings.
        :param args: Strings or lists of strings, that should match feature names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for feature in arg:
                    self._include_single_feature(feature)
            else:
                self._include_single_feature(arg)

    def _include_single_feature(self, feature):
        if type(feature) is str:
            self.included_features.add(feature)
        else:
            raise ValueError(
                "features.include() only accepts strings or lists of strings, "
                "but it received {0} which is of type '{1}'.".format(
                    str(feature), type(feature).__name__
                )
            )

    def exclude(self, *args):
        """
        Exclude one or several features of the dataset from the list of features that have been included in the
        ablation study. Features can be excluded one by one or as a list of strings. This method will check to see
        if the feature has already been included in the study before excluding it.
        :param args: Strings or lists of strings, that should match feature names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for feature in arg:
                    self._exclude_single_feature(feature)
            else:
                self._exclude_single_feature(arg)

    def _exclude_single_feature(self, feature):
        if type(feature) is str:
            if feature in self.included_features:
                self.included_features.remove(feature)
                print(
                    "Feature '{0}' is excluded from the ablation study.".format(
                        str(feature)
                    )
                )
        else:
            raise ValueError(
                "features.exclude() only accepts strings or lists of strings, "
                "but it received {0} (of type '{1}').".format(
                    str(feature), type(feature).__name__
                )
            )

    def list_all(self):
        for feature in self.included_features:
            print(feature)


class Model(object):
    def __init__(self):
        self.layers = Layers()
        self.base_model_generator = None

        # the list holding custom model generators.
        # One (extra) trial will be generated per each item in the list.
        self.custom_model_generators = []

    def set_base_model_generator(self, base_model_generator):
        self.base_model_generator = base_model_generator

    def add_custom_model_generator(self, custom_model_generator, model_identifier):
        """
        Add a custom model architecture generator, which will result in a single ablation trial.
        This method provides support for ablation study of arbitrarily complex models.

        :param custom_model_generator: A Python callable that returns a `keras.Model`. This model generator will be
        used as the model generation logic in the inner (training) loop of a single model ablation trial.
        :param model_identifier: a string that will be used to track and identify the performance of the custom model
        :type model_identifier: str
        """
        self.custom_model_generators.append((custom_model_generator, model_identifier))


class Layers(object):
    def __init__(self):
        self.included_layers = set()
        self.included_groups = set()

    def include(self, *args):
        """
        Include layers in the ablations study. Note that the first (input) and the last (output) layer of the base model
        can never be included in the ablation study.

        :param args: Strings or lists of strings, that should match layer names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for layer in arg:
                    self._include_single_layer(layer)
            else:
                self._include_single_layer(arg)

    def _include_single_layer(self, layer):
        if type(layer) is str:
            self.included_layers.add(layer)
            # print("included {}".format(layer))  # this still prints even if was duplicate
        else:
            raise ValueError(
                "layers.include() only accepts strings or lists of strings, "
                "but it received {0} which is of type '{1}'.".format(
                    str(layer), type(layer).__name__
                )
            )

    def exclude(self, *args):
        for arg in args:
            if type(arg) is list:
                for layer in arg:
                    self._exclude_single_layer(layer)
            else:
                self._exclude_single_layer(arg)

    def _exclude_single_layer(self, layer):
        if type(layer) is str:
            if layer in self.included_layers:
                self.included_layers.remove(layer)
        else:
            raise ValueError(
                "layers.exclude() only accepts strings or lists of strings, "
                "but it received {0} (of type '{1}').".format(
                    str(layer), type(layer).__name__
                )
            )

    def include_groups(self, *args, prefix=None):
        """
        Adds a group of layers that should be removed from the model together. The groups are specified either
        by being passed as a list of layer names (strings), or a string as a common prefix of their layer names.
        Each list of strings would result in a single grouping.

        :param prefix: A string that is a prefix of the names of a group of layers in the base model.
        :type prefix: str
        :param args: Lists of strings (layer names) to indicate groups of layers. The length of the list should be
        greater than one - it does not make sense to have a group of layers that consists of only one layer.
        :type args: list
        """
        if prefix is not None:
            if type(prefix) is str:
                self.included_groups.add(frozenset([prefix]))
            else:
                raise ValueError(
                    "`prefix` argument of layers.include_groups() should either be "
                    "a `NoneType` or a `str`, but it received {0} (of type '{1}'.".format(
                        str(prefix), type(prefix).__name__
                    )
                )

        for arg in args:
            if type(arg) is list and len(arg) > 1:
                self.included_groups.add(frozenset(arg))
            elif type(arg) is list and len(arg) == 1:
                raise ValueError(
                    "layers.include_groups() received a list ( {0} ) "
                    "with only one element: Did you want to include a single layer in "
                    "your ablation study? then you should use layers.include().".format(
                        str(arg)
                    )
                )
            else:
                raise ValueError(
                    "layers.include_groups() only accepts a prefix string, "
                    "or lists (with more than one element) "
                    "of strings, but it received {0} (of type '{1}').".format(
                        str(arg), type(arg).__name__
                    )
                )

    def exclude_groups(self, *args, prefix=None):
        """
        Removes a group of layers from being included in the ablation study. The groups are specified
        either by being passed as a list of layer names, or by passing a `prefix` argument, denoting a prefix shared by
        all layer names in a group.

        :param prefix: A string that is a prefix of the names of a group that is already included in the ablation study.
        :type prefix: str
        :param args: Lists of strings (layer names) that correspond to groups that are already
        included in the ablation study.
        :type args: list
        """

        if prefix is not None:
            if type(prefix) is str:
                if frozenset([prefix]) in self.included_groups:
                    self.included_groups.remove(frozenset([prefix]))
            else:
                raise ValueError(
                    "`prefix` argument of layers.exclude_groups() should either be "
                    "a `NoneType` or a `str`, but it received {0} (of type '{1}'.".format(
                        str(prefix), type(prefix).__name__
                    )
                )

        for arg in args:
            if type(arg) is list and len(arg) > 1:
                if frozenset(arg) in self.included_groups:
                    self.included_groups.remove(frozenset(arg))
            else:
                raise ValueError(
                    "layers.exclude_groups() only accepts a prefix string, or "
                    "lists (with more than one element) "
                    "of strings, but it received {0} (of type '{1}').".format(
                        str(arg), type(arg).__name__
                    )
                )

    def print_all(self):
        """Prints all single layers that are included in the current ablation study configuration."""
        if len(self.included_layers) > 0:
            print("Included single layers are: \n")
            for layer in self.included_layers:
                print(layer)
        else:
            print("There are no single layers in this ablation study configuration.")

    def print_all_groups(self):
        """Prints all layer groups that are included in the current ablation study configuration."""
        if len(self.included_groups) > 0:
            print("Included layer groups are: \n")
            for layer_group in self.included_groups:
                if len(layer_group) > 1:
                    print("--- Layer group " + str(list(layer_group)))
                elif len(layer_group) == 1:
                    print(
                        '---- All layers prefixed "' + str(list(layer_group)[0]) + '"'
                    )
        else:
            print("There are no layer groups in this ablation study configuration.")



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
import time
import random

from maggy import Searchspace


def test_searchspace_init():

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]), param2=("integer", [3, 4]))

    exp_get = [1, 5]

    assert sp.get("argument_param") == exp_get
    assert sp.argument_param == exp_get  # pylint: disable=no-member


def test_searchspace_add():

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]))

    with pytest.raises(ValueError) as excinfo:
        sp.add("argument_param", ("DOUBLE", [1, 5]))
    assert "Hyperparameter name is reserved" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add tuple with too many elements
        sp.add("param", ("DOUBLE", [1, 5], "too many"))
    assert "Hyperparameter tuple has to be of length two" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add unknown type
        sp.add("param", ("FLOAT", [1, 5]))
    assert "Hyperparameter type is not of type " in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add empty region list
        sp.add("param", ("DOUBLE", []))
    assert "Hyperparameter feasible region list" in str(excinfo.value)

    with pytest.raises(AssertionError) as excinfo:
        # add incompatible type and feasible region
        sp.add("param", ("DOUBLE", [1, 5, 5]))
        sp.add("param2", ("INTEGER", [1, 5, 5]))
    assert "For DOUBLE or " in str(excinfo.value)

    with pytest.raises(AssertionError) as excinfo:
        # lower bound higher than upper bound
        sp.add("param", ("DOUBLE", [5, 1]))
        sp.add("param2", ("INTEGER", [4, 1]))
    assert "Lower bound " in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # Non integer boundaries for integer type parameter
        sp.add("param2", ("INTEGER", [1.5, 5]))
    assert "type INTEGER need to be integer:" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # Non numeric interval boundaries
        sp.add("param2", ("DOUBLE", ["lower", 5]))
    assert "type DOUBLE need to be integer or float:" in str(excinfo.value)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
Module to encapsulate functionality related to writing to the tensorboard
log dir and programmatically structure the outputs.
"""

import tensorflow.compat.v2 as tf
from tensorboard.plugins.hparams import summary_v2 as hp
from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import summary

# __import__("tensorflow").compat.v1.enable_eager_execution()

_tensorboard_dir = None
_writer = None


def _register(trial_dir):
    global _tensorboard_dir
    global _writer
    _tensorboard_dir = trial_dir
    _writer = tf.summary.create_file_writer(_tensorboard_dir)


def logdir():
    """Returns the path to the tensorboard log directory.

    Instead of hardcoding a log dir path in a training function, users should
    make use of this function call, which will programmatically create a folder
    structure for tensorboard to visualize the machine learning experiment.

    :return: Path of the log directory in HOPSFS
    :rtype: str
    """
    global _tensorboard_dir
    return _tensorboard_dir


def _create_hparams_config(searchspace):
    hparams = []

    for key, val in searchspace.names().items():
        if val == "DOUBLE":
            hparams.append(
                hp.HParam(
                    key,
                    hp.RealInterval(
                        float(searchspace.get(key)[0]), float(searchspace.get(key)[1])
                    ),
                )
            )
        elif val == "INTEGER":
            hparams.append(
                hp.HParam(
                    key,
                    hp.IntInterval(searchspace.get(key)[0], searchspace.get(key)[1]),
                )
            )
        elif val == "DISCRETE":
            hparams.append(hp.HParam(key, hp.Discrete(searchspace.get(key))))
        elif val == "CATEGORICAL":
            hparams.append(hp.HParam(key, hp.Discrete(searchspace.get(key))))

    return hparams


def _write_hparams_config(log_dir, searchspace):
    HPARAMS = _create_hparams_config(searchspace)
    METRICS = [
        hp.Metric(
            "epoch_accuracy", group="validation", display_name="accuracy (val.)",
        ),
        hp.Metric("epoch_loss", group="validation", display_name="loss (val.)",),
        hp.Metric("epoch_accuracy", group="train", display_name="accuracy (train)",),
        hp.Metric("epoch_loss", group="train", display_name="loss (train)",),
    ]

    with tf.summary.create_file_writer(log_dir).as_default():
        hp.hparams_config(hparams=HPARAMS, metrics=METRICS)


def _write_hparams(hparams, trial_id):
    global _writer
    with _writer.as_default():
        hp.hparams(hparams, trial_id)


def _write_session_end():
    global _writer
    with _writer.as_default():
        protob = summary.session_end_pb(api_pb2.STATUS_SUCCESS)
        raw_pb = protob.SerializeToString()
        tf.summary.experimental.write_raw_pb(raw_pb, step=0)
    _writer = None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""Utility helper module for maggy experiments.
"""
import math
import os
import json
import numpy as np
from pyspark import TaskContext

from hops import util as hopsutil
from hops import hdfs as hopshdfs
from hops.experiment_impl.util import experiment_utils

from maggy import constants
from maggy.core import exceptions

DEBUG = True


def _log(msg):
    """
    Generic log function (in case logging is changed from stdout later)

    :param msg: The msg to log
    :type msg: str
    """
    if DEBUG:
        print(msg)


def num_executors(sc):
    """
    Get the number of executors configured for Jupyter

    :param sc: The SparkContext to take the executors from.
    :type sc: [SparkContext
    :return: Number of configured executors for Jupyter
    :rtype: int
    """
    sc = hopsutil._find_spark().sparkContext
    try:
        return int(sc._conf.get("spark.dynamicAllocation.maxExecutors"))
    except:  # noqa: E722
        raise RuntimeError(
            "Failed to find spark.dynamicAllocation.maxExecutors property, \
            please select your mode as either Experiment, Parallel \
            Experiments or Distributed Training."
        )


def get_partition_attempt_id():
    """Returns partitionId and attemptNumber of the task context, when invoked
    on a spark executor.
    PartitionId is ID of the RDD partition that is computed by this task.
    The first task attempt will be assigned attemptNumber = 0, and subsequent
    attempts will have increasing attempt numbers.
    Returns:
        partitionId, attemptNumber -- [description]
    """
    task_context = TaskContext.get()
    return task_context.partitionId(), task_context.attemptNumber()


def _progress_bar(done, total):

    done_ratio = done / total
    progress = math.floor(done_ratio * 30)

    bar = "["

    for i in range(30):
        if i < progress:
            bar += "="
        elif i == progress:
            bar += ">"
        else:
            bar += "."

    bar += "]"
    return bar


def json_default_numpy(obj):
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        raise TypeError(
            "Object of type {0}: {1} is not JSON serializable".format(type(obj), obj)
        )


def _finalize_experiment(
    experiment_json,
    metric,
    app_id,
    run_id,
    state,
    duration,
    logdir,
    best_logdir,
    optimization_key,
):
    """Attaches the experiment outcome as xattr metadata to the app directory.
    """
    outputs = _build_summary_json(logdir)

    if outputs:
        hopshdfs.dump(outputs, logdir + "/.summary.json")

    if best_logdir:
        experiment_json["bestDir"] = best_logdir[len(hopshdfs.project_path()) :]
    experiment_json["optimizationKey"] = optimization_key
    experiment_json["metric"] = metric
    experiment_json["state"] = state
    experiment_json["duration"] = duration
    exp_ml_id = app_id + "_" + str(run_id)
    experiment_utils._attach_experiment_xattr(exp_ml_id, experiment_json, "FULL_UPDATE")


def _build_summary_json(logdir):
    """Builds the summary json to be read by the experiments service.
    """
    combinations = []

    for trial in hopshdfs.ls(logdir):
        if hopshdfs.isdir(trial):
            return_file = trial + "/.outputs.json"
            hparams_file = trial + "/.hparams.json"
            if hopshdfs.exists(return_file) and hopshdfs.exists(hparams_file):
                metric_arr = experiment_utils._convert_return_file_to_arr(return_file)
                hparams_dict = _load_hparams(hparams_file)
                combinations.append({"parameters": hparams_dict, "outputs": metric_arr})

    return json.dumps({"combinations": combinations}, default=json_default_numpy)


def _load_hparams(hparams_file):
    """Loads the HParams configuration from a hparams file of a trial.
    """
    hparams_file_contents = hopshdfs.load(hparams_file)
    hparams = json.loads(hparams_file_contents)

    return hparams


def _handle_return_val(return_val, log_dir, optimization_key, log_file):
    """Handles the return value of the user defined training function.
    """
    experiment_utils._upload_file_output(return_val, log_dir)

    # Return type validation
    if not optimization_key:
        raise ValueError("Optimization key cannot be None.")
    if not return_val:
        raise exceptions.ReturnTypeError(optimization_key, return_val)
    if not isinstance(return_val, constants.USER_FCT.RETURN_TYPES):
        raise exceptions.ReturnTypeError(optimization_key, return_val)
    if isinstance(return_val, dict) and optimization_key not in return_val:
        raise KeyError(
            "Returned dictionary does not contain optimization key with the "
            "provided name: {}".format(optimization_key)
        )

    # validate that optimization metric is numeric
    if isinstance(return_val, dict):
        opt_val = return_val[optimization_key]
    else:
        opt_val = return_val
        return_val = {optimization_key: opt_val}

    if not isinstance(opt_val, constants.USER_FCT.NUMERIC_TYPES):
        raise exceptions.MetricTypeError(optimization_key, opt_val)

    # for key, value in return_val.items():
    #    return_val[key] = value if isinstance(value, str) else str(value)

    return_val["log"] = log_file.replace(hopshdfs.project_path(), "")

    return_file = log_dir + "/.outputs.json"
    hopshdfs.dump(json.dumps(return_val, default=json_default_numpy), return_file)

    metric_file = log_dir + "/.metric"
    hopshdfs.dump(json.dumps(opt_val, default=json_default_numpy), metric_file)

    return opt_val


def _clean_dir(clean_dir, keep=[]):
    """Deletes all files in a directory but keeps a few.
    """
    if not hopshdfs.isdir(clean_dir):
        raise ValueError(
            "{} is not a directory. Use `hops.hdfs.delete()` to delete single "
            "files.".format(clean_dir)
        )
    for path in hopshdfs.ls(clean_dir):
        if path not in keep:
            hopshdfs.delete(path, recursive=True)


def _validate_ml_id(app_id, run_id):
    """Validates if there was an experiment run previously from the same app id
    but from a different experiment (e.g. hops-util-py vs. maggy) module.
    """
    try:
        prev_ml_id = os.environ["ML_ID"]
    except KeyError:
        return app_id, run_id
    prev_app_id, _, prev_run_id = prev_ml_id.rpartition("_")
    if prev_run_id == prev_ml_id:
        # means there was no underscore found in string
        raise ValueError(
            "Found a previous ML_ID with wrong format: {}".format(prev_ml_id)
        )
    if prev_app_id == app_id and int(prev_run_id) >= run_id:
        return app_id, (int(prev_run_id) + 1)
    return app_id, run_id



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from maggy.ablation.ablator import AbstractAblator
from maggy.core.exceptions import NotSupportedError
from maggy.core.exceptions import BadArgumentsError
from hops import featurestore
from maggy.trial import Trial
import json


class LOCO(AbstractAblator):
    def __init__(self, ablation_study, final_store):
        super().__init__(ablation_study, final_store)
        self.base_dataset_generator = self.get_dataset_generator(ablated_feature=None)

    def get_number_of_trials(self):
        # the final ' + 1 ' is for the base (reference) trial with all the components
        return (
            len(self.ablation_study.features.included_features)
            + len(self.ablation_study.model.layers.included_layers)
            + len(self.ablation_study.model.layers.included_groups)
            + len(self.ablation_study.model.custom_model_generators)
            + 1
        )

    def get_dataset_generator(
        self, ablated_feature=None, dataset_type="tfrecord", shuffle_buffer_size=10000
    ):

        # for dataset generators provided by users
        if self.ablation_study.custom_dataset_generator:
            return self.ablation_study.custom_dataset_generator
        else:
            training_dataset_name = self.ablation_study.hops_training_dataset_name
            training_dataset_version = self.ablation_study.hops_training_dataset_version
            label_name = self.ablation_study.label_name

            if dataset_type == "tfrecord":

                def create_tf_dataset(num_epochs, batch_size):
                    import tensorflow as tf

                    dataset_dir = featurestore.get_training_dataset_path(
                        training_dataset_name, training_dataset_version
                    )
                    input_files = tf.io.gfile.glob(
                        (dataset_dir + "/part-r-*").replace("hopsfs", "hdfs")
                    )
                    dataset = tf.data.TFRecordDataset(input_files)
                    tf_record_schema = featurestore.get_training_dataset_tf_record_schema(
                        training_dataset_name
                    )
                    meta = featurestore.get_featurestore_metadata()
                    training_features = [
                        feature.name
                        for feature in meta.training_datasets[
                            training_dataset_name + "_" + str(training_dataset_version)
                        ].features
                    ]

                    if ablated_feature is not None:
                        training_features.remove(ablated_feature)

                    training_features.remove(label_name)

                    def decode(example_proto):
                        example = tf.io.parse_single_example(
                            example_proto, tf_record_schema
                        )
                        # prepare the features
                        x = []
                        for feature_name in training_features:
                            # temporary fix for the case of tf.int types
                            if tf_record_schema[feature_name].dtype.is_integer:
                                x.append(tf.cast(example[feature_name], tf.float32))
                            else:
                                x.append(example[feature_name])

                        # prepare the labels
                        if tf_record_schema[label_name].dtype.is_integer:
                            y = [tf.cast(example[label_name], tf.float32)]
                        else:
                            y = [example[label_name]]

                        return x, y

                    dataset = (
                        dataset.map(decode)
                        .shuffle(shuffle_buffer_size)
                        .batch(batch_size)
                        .repeat(num_epochs)
                    )
                    return dataset

                return create_tf_dataset
            else:
                raise NotSupportedError(
                    "dataset type",
                    dataset_type,
                    "Use 'tfrecord' or write your own custom dataset generator.",
                )

    def get_model_generator(self, layer_identifier=None, custom_model_generator=None):

        if layer_identifier is not None and custom_model_generator is not None:
            raise BadArgumentsError(
                "get_model_generator",
                "At least one of 'layer_identifier' or 'custom_model_generator' should be 'None'.",
            )

        # if this trial relates to a custom model, then return the provided custom model generator
        if custom_model_generator:
            return custom_model_generator[0]
        # if this is a model ablation of a base model, construct a new model generator
        # using the layer_identifier
        base_model_generator = self.ablation_study.model.base_model_generator
        if layer_identifier is None:
            return base_model_generator

        def model_generator():
            import tensorflow as tf

            base_model = base_model_generator()

            list_of_layers = [
                base_layer for base_layer in base_model.get_config()["layers"]
            ]
            if type(layer_identifier) is str:
                # ablation of a single layer
                for base_layer in reversed(list_of_layers[1:-1]):
                    # the first (input) and last (output) layers should not be considered, hence list_of_layers[1:-1]
                    if base_layer["config"]["name"] == layer_identifier:
                        list_of_layers.remove(base_layer)
                        break
            elif type(layer_identifier) is set:
                # ablation of a layer group - all the layers in the group should be removed together
                if len(layer_identifier) > 1:
                    # group of layers (non-prefix)
                    for base_layer in reversed(list_of_layers[1:-1]):
                        if base_layer["config"]["name"] in layer_identifier:
                            list_of_layers.remove(base_layer)
                elif len(layer_identifier) == 1:
                    # layer_identifier is a prefix
                    prefix = list(layer_identifier)[0].lower()
                    for base_layer in reversed(list_of_layers[1:-1]):
                        if base_layer["config"]["name"].lower().startswith(prefix):
                            list_of_layers.remove(base_layer)

            base_json = base_model.to_json()
            new_dict = json.loads(base_json)
            new_dict["config"]["layers"] = list_of_layers
            new_json = json.dumps(new_dict)
            new_model = tf.keras.models.model_from_json(new_json)

            return new_model

        return model_generator

    def initialize(self):
        """
        Prepares all the trials for LOCO policy (Leave One Component Out).
        In total `n+1` trials will be generated where `n` is equal to the number of components
        (e.g. features and layers) that are included in the ablation study
        (i.e. the components that will be removed one-at-a-time). The first trial will include all the components and
        can be regarded as the base for comparison.
        """

        # 0 - add first trial with all the components (base/reference trial)
        self.trial_buffer.append(
            Trial(self.create_trial_dict(None, None), trial_type="ablation")
        )

        # generate remaining trials based on the ablation study configuration:
        # 1 - generate feature ablation trials
        for feature in self.ablation_study.features.included_features:
            self.trial_buffer.append(
                Trial(
                    self.create_trial_dict(ablated_feature=feature),
                    trial_type="ablation",
                )
            )

        # 2 - generate single-layer ablation trials
        for layer in self.ablation_study.model.layers.included_layers:
            self.trial_buffer.append(
                Trial(
                    self.create_trial_dict(layer_identifier=layer),
                    trial_type="ablation",
                )
            )

        # 3 - generate layer-groups ablation trials
        # each element of `included_groups` is a frozenset of a set, so we cast again to get a set
        # why frozensets in the first place? because a set can only contain immutable elements
        # hence elements (layer group identifiers) are frozensets

        for layer_group in self.ablation_study.model.layers.included_groups:
            self.trial_buffer.append(
                Trial(
                    self.create_trial_dict(layer_identifier=set(layer_group)),
                    trial_type="ablation",
                )
            )

        # 4 - generate ablation trials based on custom model generators

        for custom_model_generator in self.ablation_study.model.custom_model_generators:
            self.trial_buffer.append(
                Trial(
                    self.create_trial_dict(
                        custom_model_generator=custom_model_generator
                    ),
                    trial_type="ablation",
                )
            )

    def get_trial(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):
        return

    def create_trial_dict(
        self, ablated_feature=None, layer_identifier=None, custom_model_generator=None
    ):
        """
        Creates a trial dictionary that can be used for creating a Trial instance.

        :param ablated_feature: a string representing the name of a feature, or None
        :param layer_identifier: A string representing the name of a single layer, or a set representing a layer group.
        If the set has only one element, it is regarded as a prefix, so all layers with that prefix in their names
        would be regarded as a layer group. Otherwise, if the set has more than one element, the layers with
        names corresponding to those elements would be regarded as a layer group.
        :return: A trial dictionary that can be passed to maggy.Trial() constructor.
        :rtype: dict
        """

        trial_dict = {}

        # 1 - determine the dataset generation logic
        if ablated_feature is None:
            trial_dict["dataset_function"] = self.base_dataset_generator
            trial_dict["ablated_feature"] = "None"
        else:
            trial_dict["dataset_function"] = self.get_dataset_generator(
                ablated_feature, dataset_type="tfrecord"
            )
            trial_dict["ablated_feature"] = ablated_feature

        # 2 - determine the model generation logic
        # 2.1 - no model ablation

        if layer_identifier is None and custom_model_generator is None:
            trial_dict[
                "model_function"
            ] = self.ablation_study.model.base_model_generator
            trial_dict["ablated_layer"] = "None"
        # 2.2 - layer ablation based on base model generator
        elif layer_identifier is not None and custom_model_generator is None:
            trial_dict["model_function"] = self.get_model_generator(
                layer_identifier=layer_identifier
            )
            if type(layer_identifier) is str:
                trial_dict["ablated_layer"] = layer_identifier
            elif type(layer_identifier) is set:
                if len(layer_identifier) > 1:
                    trial_dict["ablated_layer"] = str(list(layer_identifier))
                elif len(layer_identifier) == 1:
                    trial_dict["ablated_layer"] = "Layers prefixed " + str(
                        list(layer_identifier)[0]
                    )
        # 2.3 - model ablation based on a custom model generator
        elif layer_identifier is None and custom_model_generator is not None:
            trial_dict["model_function"] = self.get_model_generator(
                custom_model_generator
            )
            trial_dict["ablated_layer"] = "Custom model: " + custom_model_generator[1]

        return trial_dict



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from abc import ABC, abstractmethod


class AbstractEarlyStop(ABC):
    """An Abstract class to implement custom early stopping criteria."""

    @staticmethod
    @abstractmethod
    def earlystop_check(to_check, finalized_trials, direction):
        """A abstract static method that needs to be implemented with a custom
        early stopping criterium.

        The function is called internally in the user specified interval
        with three arguments. It is necessary to add these to the function
        definition.

        :param to_check: A dictionary of currently running
        trials, where the key is the `trial_id` and values are Trial objects.
        :type to_check: dictionary
        :param finalized_trials: A list of finalized Trial objects.
        :type finalized_trials: list
        :param direction: A string describing the search objective, i.e. 'min'
        or 'max'.
        :type direction: str
        """
        pass



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from abc import ABC, abstractmethod


class AbstractOptimizer(ABC):
    def __init__(self):
        self.searchspace = None
        self.num_trials = None
        self.trial_store = None
        self.final_store = None
        self.direction = None

    @abstractmethod
    def initialize(self):
        """
        A hook for the developer to initialize the optimizer.
        """
        pass

    @abstractmethod
    def get_suggestion(self, trial=None):
        """
        Return a `Trial` to be assigned to an executor, or `None` if there are
        no trials remaining in the experiment.

        :rtype: `Trial` or `None`
        """
        pass

    @abstractmethod
    def finalize_experiment(self, trials):
        """
        This method will be called before finishing the experiment. Developers
        can implement this method e.g. for cleanup or extra logging.
        """
        pass

    def name(self):
        return str(self.__class__.__name__)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
API Module for the user to include in his training code.

"""
import threading
from datetime import datetime

from hops import hdfs as hopshdfs

from maggy import constants
from maggy.core import exceptions


class Reporter(object):
    """
    Thread-safe store for sending a metric and logs from executor to driver
    """

    def __init__(self, log_file, partition_id, task_attempt, print_executor):
        self.metric = None
        self.step = -1
        self.lock = threading.RLock()
        self.stop = False
        self.trial_id = None
        self.trial_log_file = None
        self.logs = ""
        self.log_file = log_file
        self.partition_id = partition_id
        self.task_attempt = task_attempt
        self.print_executor = print_executor

        # Open executor log file descriptor
        # This log is for all maggy system related log messages
        if not hopshdfs.exists(log_file):
            hopshdfs.dump("", log_file)
        self.fd = hopshdfs.open_file(log_file, flags="w")
        self.trial_fd = None

    def init_logger(self, trial_log_file):
        """Initializes the trial log file
        """
        self.trial_log_file = trial_log_file
        # Open trial log file descriptor
        if not hopshdfs.exists(self.trial_log_file):
            hopshdfs.dump("", self.trial_log_file)
        self.trial_fd = hopshdfs.open_file(self.trial_log_file, flags="w")

    def close_logger(self):
        """Savely closes the file descriptors of the log files.

        close() can be called multiple times and flushes the buffer contents
        before closing
        """
        with self.lock:
            if self.trial_fd:
                self.trial_fd.close()
            self.fd.close()

    # report
    def broadcast(self, metric, step=None):
        """Broadcast a metric to the experiment driver with the heartbeat.

        :param metric: Metric to be broadcasted
        :type metric: int, float
        :param step: The iteration step which produced the metric, e.g. batch or
            epoch number, or any other monotonically increasing progress attribute
        :type step: int
        :raises exception: EarlyStopException if told by the experiment driver
        """
        with self.lock:
            # if stop == True -> raise exception to break training function
            if step is None:
                step = self.step + 1
            if not isinstance(metric, constants.USER_FCT.NUMERIC_TYPES):
                raise exceptions.BroadcastMetricTypeError(metric)
            elif not isinstance(step, constants.USER_FCT.NUMERIC_TYPES):
                raise exceptions.BroadcastStepTypeError(metric, step)
            elif step < self.step:
                raise exceptions.BroadcastStepValueError(metric, step, self.step)
            else:
                self.step = step
                self.metric = metric
            if self.stop:
                raise exceptions.EarlyStopException(metric)

    def log(self, log_msg, jupyter=False):
        """Logs a message to the executor logfile and executor stderr and
        optionally prints the message in jupyter.

        :param log_msg: Message to log.
        :type log_msg: str
        :param verbose: Print in Jupyter Notebook, defaults to True
        :type verbose: bool, optional
        """
        with self.lock:
            try:
                msg = (datetime.now().isoformat() + " ({0}/{1}): {2} \n").format(
                    self.partition_id, self.task_attempt, log_msg
                )
                if jupyter:
                    jupyter_log = str(self.partition_id) + ": " + log_msg
                    self.trial_fd.write(msg.encode())
                    self.logs = self.logs + jupyter_log + "\n"
                else:
                    self.fd.write(msg.encode())
                    if self.trial_fd:
                        self.trial_fd.write(msg.encode())
                    self.print_executor(msg)
            # Throws ValueError when operating on closed HDFS file object
            # Throws AttributeError when calling file ops on NoneType object
            except (IOError, ValueError, AttributeError) as e:
                self.fd.write(
                    ("An error occurred while writing logs: {}".format(e)).encode()
                )

    def get_data(self):
        """Returns the metric and logs to be sent to the experiment driver.
        """
        with self.lock:
            log_to_send = self.logs
            self.logs = ""
            return self.metric, self.step, log_to_send

    def reset(self):
        """
        Resets the reporter to the initial state in order to start a new
        trial.
        """
        with self.lock:
            self.metric = None
            self.step = -1
            self.stop = False
            self.trial_id = None
            self.fd.flush()
            self.trial_fd.close()
            self.trial_fd = None
            self.trial_log_file = None

    def early_stop(self):
        with self.lock:
            if self.metric is not None:
                self.stop = True

    def get_trial_id(self):
        with self.lock:
            return self.trial_id

    def set_trial_id(self, trial_id):
        with self.lock:
            self.trial_id = trial_id



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from maggy.optimizer.abstractoptimizer import AbstractOptimizer
from maggy.searchspace import Searchspace
from maggy.trial import Trial


class RandomSearch(AbstractOptimizer):
    def __init__(self):
        super().__init__()
        self.trial_buffer = []

    def initialize(self):

        if (
            Searchspace.DOUBLE not in self.searchspace.names().values()
            and Searchspace.INTEGER not in self.searchspace.names().values()
        ):
            raise NotImplementedError(
                "Searchspace needs at least one continuous parameter for random search."
            )

        list_of_random_trials = self.searchspace.get_random_parameter_values(
            self.num_trials
        )
        for parameters_dict in list_of_random_trials:
            self.trial_buffer.append(Trial(parameters_dict, trial_type="optimization"))

    def get_suggestion(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):
        return



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import statistics
from maggy.earlystop.abstractearlystop import AbstractEarlyStop


class MedianStoppingRule(AbstractEarlyStop):
    """The Median Stopping Rule implements the simple strategy of stopping a
    trial if its performance falls below the median of other trials at similar
    points in time.
    """

    @staticmethod
    def earlystop_check(to_check, finalized_trials, direction):

        results = []
        median = None

        # count step from zero so it can be used as index for array
        step = len(to_check.metric_history)

        if step > 0:

            for fin_trial in finalized_trials:

                if len(fin_trial.metric_history) >= step:
                    avg = sum(fin_trial.metric_history[:step]) / float(step)
                    results.append(avg)

            try:
                median = statistics.median(results)
            except statistics.StatisticsError as e:
                raise Exception(
                    "Warning: StatisticsError when calling early stop method\n{}".format(
                        e
                    )
                )

            if median is not None:
                if direction == "max":
                    if max(to_check.metric_history) < median:
                        return to_check.trial_id
                elif direction == "min":
                    if min(to_check.metric_history) > median:
                        return to_check.trial_id
            return None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import tensorflow as tf


class KerasBatchEnd(tf.keras.callbacks.Callback):
    """A Keras callback reporting a specified `metric` at the end of the batch
    to the maggy experiment driver.

    `loss` is always available as a metric, and optionally `acc` (if accuracy
    monitoring is enabled, that is, accuracy is added to keras model metrics).
    Validation metrics are not available for the BatchEnd callback. Validation
    after every batch would be too expensive.
    Default is training loss (`loss`).

    Example usage:

    >>> from maggy.callbacks import KerasBatchEnd
    >>> callbacks = [KerasBatchEnd(reporter, metric='acc')]
    """

    def __init__(self, reporter, metric="loss"):
        super().__init__()
        self.metric_name = metric
        self.reporter = reporter

    def on_batch_end(self, batch, logs={}):
        self.reporter.broadcast(logs.get(self.metric_name, 0))


class KerasEpochEnd(tf.keras.callbacks.Callback):
    """A Keras callback reporting a specified `metric` at the end of an epoch
    to the maggy experiment driver.

    `val_loss` is always available as a metric, and optionally `val_acc` (if
    accuracy monitoring is enabled, that is, accuracy is added to keras model
    metrics). Training metrics are available under the names `loss` and `acc`.
    Default is validation loss (`val_loss`).

    Example usage:

    >>> from maggy.callbacks import KerasBatchEnd
    >>> callbacks = [KerasBatchEnd(reporter, metric='val_acc')]
    """

    def __init__(self, reporter, metric="val_loss"):
        super().__init__()
        self.metric_name = metric
        self.reporter = reporter

    def on_epoch_end(self, epoch, logs={}):
        self.reporter.broadcast(logs.get(self.metric_name, 0), epoch)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
Maggy specific exceptions.
"""


class EarlyStopException(Exception):
    """Raised by the reporter when a early stop signal is received."""

    def __init__(self, metric):
        super().__init__()
        self.metric = metric


class NotSupportedError(Exception):
    """Raised when we are dealing with a situation that we do not (yet)
    support, e.g., a specific dataset type.
    """

    def __init__(self, category, value, suggestion=""):
        self.message = "({0}: {1}) is not supported by Maggy.{2}".format(
            category, value, suggestion
        )
        super().__init__(self.message)


class ReturnTypeError(TypeError):
    """User defined training function returns value of wrong type."""

    def __init__(self, optimization_key, return_type):
        self.message = (
            "Training function cannot return value of type: {}. "
            "Return single numeric value or 'dict' containing optimization key"
            " `{}` with numeric value".format(
                type(return_type).__name__, optimization_key
            )
        )
        super().__init__(self.message)


class MetricTypeError(TypeError):
    """User defined training function returns metric of wrong type."""

    def __init__(self, optimization_key, return_type):
        self.message = (
            "The optimization metric `{}` returned by the training function is"
            " of type: {}. The optimization metric can only be numeric".format(
                optimization_key, type(return_type).__name__
            )
        )
        super().__init__(self.message)


class BroadcastMetricTypeError(TypeError):
    """User defined training function broadcasts metric of wrong type."""

    def __init__(self, return_type):
        self.message = (
            "The optimization metric broadcast by the training function with "
            "the reporter is of type: {}. The optimization metric can only "
            "be numeric".format(type(return_type).__name__)
        )
        super().__init__(self.message)


class BroadcastStepTypeError(TypeError):
    """User defined training function broadcasts metric with a non-numeric step
    type.
    """

    def __init__(self, value, step):
        self.message = (
            "The optimization metric `{}` was broadcast by the training "
            " function in step {}, which is of type {}. The step value can "
            "only be numeric.".format(value, step, type(value).__name__)
        )
        super().__init__(self.message)


class BroadcastStepValueError(ValueError):
    """User defined training function broadcasts metric with a
    non-monotonically increasing step attribute.
    """

    def __init__(self, value, step, prev_step):
        self.message = (
            "The optimization metric `{}` was broadcast by the training "
            " function in step {}, while the previous step was {}. The steps "
            "should be a monotonically increasing attribute.".format(
                value, step, prev_step
            )
        )
        super().__init__(self.message)


class BadArgumentsError(Exception):
    """Raised when a function or method has been called with incompatible arguments.
    This can be used by developers to prevent bad usage of their functions
    or classes by other developers.
    """

    def __init__(self, callable, suggestion=""):
        self.message = "{0} was called using incompatible arguments. {1}".format(
            callable, suggestion
        )
        super().__init__(self.message)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
The experiment driver implements the functionality for scheduling trials on
maggy.
"""
import queue
import threading
import json
import os
import secrets
import time
from datetime import datetime

from hops import constants as hopsconstants
from hops import hdfs as hopshdfs
from hops import util as hopsutil
from hops.experiment_impl.util import experiment_utils

from maggy import util
from maggy.optimizer import AbstractOptimizer, RandomSearch, Asha, SingleRun
from maggy.core import rpc
from maggy.trial import Trial
from maggy.earlystop import AbstractEarlyStop, MedianStoppingRule, NoStoppingRule
from maggy.searchspace import Searchspace

from maggy.ablation.ablator import AbstractAblator
from maggy.ablation.ablator.loco import LOCO
from maggy.ablation.ablationstudy import AblationStudy

driver_secret = None


class ExperimentDriver(object):

    SECRET_BYTES = 8

    # @Moritz:
    # for now, we infer the experiment type (an optimization experiment or an ablation study)
    # using keyword arguments, and set self.experiment_type to an according string.
    # Some of these arguments are required for any maggy experiment:
    # num_trials, name, num_executors, hb_interval, description, app_dir, log_dir, trial_dir
    # while some are specific to the type of experiment. For example, if the ExperimentDriver constructor
    # is called with a `searchspace` parameter, we infer that it is a hyperparameter optimization task,
    # and if it is called with `ablationstudy` parameter, we infer that it's an ablation study.

    def __init__(self, experiment_type, **kwargs):

        global driver_secret

        # COMMON EXPERIMENT SETUP
        self._final_store = []
        self._trial_store = {}
        self.num_executors = kwargs.get("num_executors")
        self._message_q = queue.Queue()
        self.name = kwargs.get("name")
        self.experiment_done = False
        self.worker_done = False
        self.hb_interval = kwargs.get("hb_interval")
        self.description = kwargs.get("description")
        self.experiment_type = experiment_type
        self.es_interval = kwargs.get("es_interval")
        self.es_min = kwargs.get("es_min")

        # TYPE-SPECIFIC EXPERIMENT SETUP
        if self.experiment_type == "optimization":
            # set up an optimization experiment

            self.num_trials = kwargs.get("num_trials", 1)

            searchspace = kwargs.get("searchspace")
            if isinstance(searchspace, Searchspace):
                self.searchspace = searchspace
            elif searchspace is None:
                self.searchspace = Searchspace()
            else:
                raise Exception(
                    "The experiment's search space should be an instance of maggy.Searchspace, "
                    "but it is {0} (of type '{1}').".format(
                        str(searchspace), type(searchspace).__name__
                    )
                )

            optimizer = kwargs.get("optimizer")

            if optimizer is None:
                if len(self.searchspace.names()) == 0:
                    self.optimizer = SingleRun()
                else:
                    raise Exception(
                        "Searchspace has to be empty or None to use without optimizer"
                    )
            elif isinstance(optimizer, str):
                if optimizer.lower() == "randomsearch":
                    self.optimizer = RandomSearch()
                elif optimizer.lower() == "asha":
                    self.optimizer = Asha()
                elif optimizer.lower() == "none":
                    if len(self.searchspace.names()) == 0:
                        self.optimizer = SingleRun()
                    else:
                        raise Exception(
                            "Searchspace has to be empty or None to use without Optimizer."
                        )
                else:
                    raise Exception(
                        "Unknown Optimizer. Can't initialize experiment driver."
                    )
            elif isinstance(optimizer, AbstractOptimizer):
                self.optimizer = optimizer
                print("Custom Optimizer initialized.")
            else:
                raise Exception(
                    "The experiment's optimizer should either be an string indicating the name "
                    "of an implemented optimizer (such as 'randomsearch') or an instance of "
                    "maggy.optimizer.AbstractOptimizer, "
                    "but it is {0} (of type '{1}').".format(
                        str(optimizer), type(optimizer).__name__
                    )
                )

            direction = kwargs.get("direction", "max")
            if isinstance(direction, str) and direction.lower() in ["min", "max"]:
                self.direction = direction.lower()
            else:
                raise Exception(
                    "The experiment's direction should be an string (either 'min' or 'max') "
                    "but it is {0} (of type '{1}').".format(
                        str(direction), type(direction).__name__
                    )
                )

            es_policy = kwargs.get("es_policy")
            if isinstance(es_policy, str):
                if es_policy.lower() == "median":
                    self.earlystop_check = MedianStoppingRule.earlystop_check
                elif es_policy.lower() == "none":
                    self.earlystop_check = NoStoppingRule.earlystop_check
                else:
                    raise Exception(
                        "The experiment's early stopping policy should either be a string ('median' or 'none') "
                        "or a custom policy that is an instance of maggy.earlystop.AbstractEarlyStop, "
                        "but it is {0} (of type '{1}').".format(
                            str(es_policy), type(es_policy).__name__
                        )
                    )
            elif isinstance(es_policy, AbstractEarlyStop):
                self.earlystop_check = es_policy.earlystop_check
                print("Custom Early Stopping policy initialized.")
            else:
                raise Exception(
                    "The experiment's early stopping policy should either be a string ('median' or 'none') "
                    "or a custom policy that is an instance of maggy.earlystop.AbstractEarlyStop, "
                    "but it is {0} (of type '{1}').".format(
                        str(es_policy), type(es_policy).__name__
                    )
                )

            self.es_interval = kwargs.get("es_interval")
            self.es_min = kwargs.get("es_min")

            self.result = {"best_val": "n.a.", "num_trials": 0, "early_stopped": 0}

        elif self.experiment_type == "ablation":
            # set up an ablation study experiment
            self.earlystop_check = NoStoppingRule.earlystop_check

            ablation_study = kwargs.get("ablation_study")
            if isinstance(ablation_study, AblationStudy):
                self.ablation_study = ablation_study
            else:
                raise Exception(
                    "The experiment's ablation study configuration should be an instance of "
                    "maggy.ablation.AblationStudy, "
                    "but it is {0} (of type '{1}').".format(
                        str(ablation_study), type(ablation_study).__name__
                    )
                )

            searchspace = kwargs.get("searchspace")
            if not searchspace:
                self.searchspace = Searchspace()
            else:
                raise Exception(
                    "The experiment's search space should be None for ablation experiments, "
                    "but it is {0} (of type '{1}').".format(
                        str(searchspace), type(searchspace).__name__
                    )
                )

            ablator = kwargs.get("ablator")
            if isinstance(ablator, str):
                if ablator.lower() == "loco":
                    self.ablator = LOCO(ablation_study, self._final_store)
                    self.num_trials = self.ablator.get_number_of_trials()
                    if self.num_executors > self.num_trials:
                        self.num_executors = self.num_trials
                else:
                    raise Exception(
                        "The experiment's ablation study policy should either be a string ('loco') "
                        "or a custom policy that is an instance of maggy.ablation.ablation.AbstractAblator, "
                        "but it is {0} (of type '{1}').".format(
                            str(ablator), type(ablator).__name__
                        )
                    )
            elif isinstance(ablator, AbstractAblator):
                self.ablator = ablator
                print("Custom Ablator initialized. \n")
            else:
                raise Exception(
                    "The experiment's ablation study policy should either be a string ('loco') "
                    "or a custom policy that is an instance of maggy.ablation.ablation.AbstractAblator, "
                    "but it is {0} (of type '{1}').".format(
                        str(ablator), type(ablator).__name__
                    )
                )

            self.result = {"best_val": "n.a.", "num_trials": 0, "early_stopped": "n.a"}
        else:
            raise Exception(
                "Unknown experiment type. experiment_type should be either 'optimization' or 'ablation', "
                "but it is {0}.".format(str(self.experiment_type))
            )

        # FINALIZE EXPERIMENT SETUP
        self.server = rpc.Server(self.num_executors)
        if not driver_secret:
            driver_secret = self._generate_secret(ExperimentDriver.SECRET_BYTES)
        self._secret = driver_secret
        self.job_start = datetime.now()
        self.executor_logs = ""
        self.maggy_log = ""
        self.log_lock = threading.RLock()
        self.log_file = kwargs.get("log_dir") + "/maggy.log"
        self.log_dir = kwargs.get("log_dir")
        self.exception = None

        # Open File desc for HDFS to log
        if not hopshdfs.exists(self.log_file):
            hopshdfs.dump("", self.log_file)
        self.fd = hopshdfs.open_file(self.log_file, flags="w")

    def init(self, job_start):

        self.server_addr = self.server.start(self)

        self.job_start = job_start

        if self.experiment_type == "optimization":
            # Set references to data in optimizer
            self.optimizer.num_trials = self.num_trials
            self.optimizer.searchspace = self.searchspace
            self.optimizer.trial_store = self._trial_store
            self.optimizer.final_store = self._final_store
            self.optimizer.direction = self.direction
            self.optimizer.initialize()
        elif self.experiment_type == "ablation":
            # set references to data in ablator
            self.ablator.ablation_study = self.ablation_study
            self.ablator.final_store = self._final_store
            self.ablator.initialize()

        self._start_worker()

    def finalize(self, job_end):

        results = ""

        if self.experiment_type == "optimization":

            _ = self.optimizer.finalize_experiment(self._final_store)

            self.job_end = job_end

            self.duration = experiment_utils._seconds_to_milliseconds(
                self.job_end - self.job_start
            )

            self.duration_str = experiment_utils._time_diff(
                self.job_start, self.job_end
            )

            results = (
                "\n------ "
                + self.optimizer.name()
                + " Results ------ direction("
                + self.direction
                + ") \n"
                "BEST combination "
                + json.dumps(self.result["best_hp"])
                + " -- metric "
                + str(self.result["best_val"])
                + "\n"
                "WORST combination "
                + json.dumps(self.result["worst_hp"])
                + " -- metric "
                + str(self.result["worst_val"])
                + "\n"
                "AVERAGE metric -- " + str(self.result["avg"]) + "\n"
                "EARLY STOPPED Trials -- " + str(self.result["early_stopped"]) + "\n"
                "Total job time " + self.duration_str + "\n"
            )

        elif self.experiment_type == "ablation":

            _ = self.ablator.finalize_experiment(self._final_store)
            self.job_end = job_end

            self.duration = experiment_utils._seconds_to_milliseconds(
                self.job_end - self.job_start
            )

            self.duration_str = experiment_utils._time_diff(
                self.job_start, self.job_end
            )

            results = (
                "\n------ "
                + self.ablator.name()
                + " Results ------ \n"
                + "BEST Config Excludes "
                + json.dumps(self.result["best_config"])
                + " -- metric "
                + str(self.result["best_val"])
                + "\n"
                + "WORST Config Excludes "
                + json.dumps(self.result["worst_config"])
                + " -- metric "
                + str(self.result["worst_val"])
                + "\n"
                + "AVERAGE metric -- "
                + str(self.result["avg"])
                + "\n"
                + "Total Job Time "
                + self.duration_str
                + "\n"
            )

        print(results)

        self._log(results)

        hopshdfs.dump(
            json.dumps(self.result, default=util.json_default_numpy),
            self.log_dir + "/result.json",
        )
        sc = hopsutil._find_spark().sparkContext
        hopshdfs.dump(self.json(sc), self.log_dir + "/maggy.json")

        return self.result

    def get_trial(self, trial_id):
        return self._trial_store[trial_id]

    def add_trial(self, trial):
        self._trial_store[trial.trial_id] = trial

    def add_message(self, msg):
        self._message_q.put(msg)

    def _start_worker(self):
        def _target_function(self):

            try:
                while not self.worker_done:
                    trial = None
                    # get a message
                    try:
                        msg = self._message_q.get_nowait()
                    except queue.Empty:
                        msg = {"type": None}

                    # depending on message do the work
                    # 1. METRIC
                    if msg["type"] == "METRIC":
                        # append executor logs if in the message
                        logs = msg.get("logs", None)
                        if logs is not None:
                            with self.log_lock:
                                self.executor_logs = self.executor_logs + logs

                        step = None
                        if msg["trial_id"] is not None and msg["data"] is not None:
                            step = self.get_trial(msg["trial_id"]).append_metric(
                                msg["data"]
                            )

                        # maybe these if statements should be in a function
                        # also this could be made a separate message
                        # i.e. step nr is added to the queue as message which will
                        # then later be checked for early stopping, just to not
                        # block for too long for other messages
                        if self.earlystop_check != NoStoppingRule.earlystop_check:
                            if len(self._final_store) > self.es_min:
                                if step is not None and step != 0:
                                    if step % self.es_interval == 0:
                                        try:
                                            to_stop = self.earlystop_check(
                                                self.get_trial(msg["trial_id"]),
                                                self._final_store,
                                                self.direction,
                                            )
                                        except Exception as e:
                                            self._log(e)
                                            to_stop = None
                                        if to_stop is not None:
                                            self._log(
                                                "Trials to stop: {}".format(to_stop)
                                            )
                                            self.get_trial(to_stop).set_early_stop()

                        # 2. BLACKLIST the trial
                    elif msg["type"] == "BLACK":
                        trial = self.get_trial(msg["trial_id"])
                        with trial.lock:
                            trial.status = Trial.SCHEDULED
                            self.server.reservations.assign_trial(
                                msg["partition_id"], msg["trial_id"]
                            )

                    # 3. FINAL
                    elif msg["type"] == "FINAL":
                        # set status
                        # get trial only once
                        trial = self.get_trial(msg["trial_id"])

                        logs = msg.get("logs", None)
                        if logs is not None:
                            with self.log_lock:
                                self.executor_logs = self.executor_logs + logs

                        # finalize the trial object
                        with trial.lock:
                            trial.status = Trial.FINALIZED
                            trial.final_metric = msg["data"]
                            trial.duration = experiment_utils._seconds_to_milliseconds(
                                time.time() - trial.start
                            )

                        # move trial to the finalized ones
                        self._final_store.append(trial)
                        self._trial_store.pop(trial.trial_id)

                        # update result dictionary
                        self._update_result(trial)
                        # keep for later in case tqdm doesn't work
                        self.maggy_log = self._update_maggy_log()
                        self._log(self.maggy_log)

                        hopshdfs.dump(
                            trial.to_json(),
                            self.log_dir + "/" + trial.trial_id + "/trial.json",
                        )

                        # assign new trial
                        if self.experiment_type == "optimization":
                            trial = self.optimizer.get_suggestion(trial)
                        elif self.experiment_type == "ablation":
                            trial = self.ablator.get_trial(trial)
                        if trial is None:
                            self.server.reservations.assign_trial(
                                msg["partition_id"], None
                            )
                            self.experiment_done = True
                        elif trial == "IDLE":
                            self.add_message(
                                {
                                    "type": "IDLE",
                                    "partition_id": msg["partition_id"],
                                    "idle_start": time.time(),
                                }
                            )
                            self.server.reservations.assign_trial(
                                msg["partition_id"], None
                            )
                        else:
                            with trial.lock:
                                trial.start = time.time()
                                trial.status = Trial.SCHEDULED
                                self.server.reservations.assign_trial(
                                    msg["partition_id"], trial.trial_id
                                )
                                self.add_trial(trial)

                    # 4. Let executor be idle
                    elif msg["type"] == "IDLE":
                        # execute only every 0.1 seconds but do not block thread
                        if (
                            self.experiment_type == "optimization"
                            and time.time() - msg["idle_start"] > 0.1
                        ):
                            trial = self.optimizer.get_suggestion()
                            if trial is None:
                                self.server.reservations.assign_trial(
                                    msg["partition_id"], None
                                )
                                self.experiment_done = True
                            elif trial == "IDLE":
                                # reset timeout
                                msg["idle_start"] = time.time()
                                self.add_message(msg)
                            else:
                                with trial.lock:
                                    trial.start = time.time()
                                    trial.status = Trial.SCHEDULED
                                    self.server.reservations.assign_trial(
                                        msg["partition_id"], trial.trial_id
                                    )
                                    self.add_trial(trial)
                        elif self.experiment_type == "optimization":
                            self.add_message(msg)

                    # 4. REG
                    elif msg["type"] == "REG":
                        if self.experiment_type == "optimization":
                            trial = self.optimizer.get_suggestion()
                        elif self.experiment_type == "ablation":
                            trial = self.ablator.get_trial()
                        if trial is None:
                            self.server.reservations.assign_trial(
                                msg["partition_id"], None
                            )
                            self.experiment_done = True
                        elif trial == "IDLE":
                            # reset timeout
                            msg["idle_start"] = time.time()
                            self.add_message(msg)
                        else:
                            with trial.lock:
                                trial.start = time.time()
                                trial.status = Trial.SCHEDULED
                                self.server.reservations.assign_trial(
                                    msg["partition_id"], trial.trial_id
                                )
                                self.add_trial(trial)
            except Exception as exc:
                # Exception can't be propagated to parent thread
                # therefore log the exception and fail experiment
                self._log(exc)
                self.exception = exc
                self.server.stop()

        t = threading.Thread(target=_target_function, args=(self,))
        t.daemon = True
        t.start()

    def stop(self):
        """Stop the Driver's worker thread and server."""
        self.worker_done = True
        self.server.stop()
        self.fd.flush()
        self.fd.close()

    def json(self, sc):
        """Get all relevant experiment information in JSON format.
        """
        user = None
        if hopsconstants.ENV_VARIABLES.HOPSWORKS_USER_ENV_VAR in os.environ:
            user = os.environ[hopsconstants.ENV_VARIABLES.HOPSWORKS_USER_ENV_VAR]

        experiment_json = {
            "project": hopshdfs.project_name(),
            "user": user,
            "name": self.name,
            "module": "maggy",
            "app_id": str(sc.applicationId),
            "start": time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(self.job_start)),
            "memory_per_executor": str(sc._conf.get("spark.executor.memory")),
            "gpus_per_executor": str(sc._conf.get("spark.executor.gpus")),
            "executors": self.num_executors,
            "logdir": self.log_dir,
            # 'versioned_resources': versioned_resources,
            "description": self.description,
            "experiment_type": self.experiment_type,
        }

        if self.experiment_type == "optimization":
            experiment_json["hyperparameter_space"] = json.dumps(
                self.searchspace.to_dict()
            )
            experiment_json["function"] = self.optimizer.name()
        elif self.experiment_type == "ablation":
            experiment_json["ablation_study"] = json.dumps(
                self.ablation_study.to_dict()
            )
            experiment_json["ablator"] = self.ablator.name()

        if self.experiment_done:
            experiment_json["status"] = "FINISHED"
            experiment_json["finished"] = time.strftime(
                "%Y-%m-%dT%H:%M:%S", time.localtime(self.job_end)
            )
            experiment_json["duration"] = self.duration
            if self.experiment_type == "optimization":
                experiment_json["hyperparameter"] = json.dumps(self.result["best_hp"])
            experiment_json["metric"] = self.result["best_val"]

        else:
            experiment_json["status"] = "RUNNING"

        return json.dumps(experiment_json, default=util.json_default_numpy)

    def _generate_secret(self, nbytes):
        """Generates a secret to be used by all clients during the experiment
        to authenticate their messages with the experiment driver.
        """
        return secrets.token_hex(nbytes=nbytes)

    def _update_result(self, trial):
        """Given a finalized trial updates the current result's best and
        worst trial.
        """

        metric = trial.final_metric
        param_string = trial.params
        trial_id = trial.trial_id

        if self.experiment_type == "optimization":
            # First finalized trial
            if self.result.get("best_id", None) is None:
                self.result = {
                    "best_id": trial_id,
                    "best_val": metric,
                    "best_hp": param_string,
                    "worst_id": trial_id,
                    "worst_val": metric,
                    "worst_hp": param_string,
                    "avg": metric,
                    "metric_list": [metric],
                    "num_trials": 1,
                    "early_stopped": 0,
                }

                if trial.early_stop:
                    self.result["early_stopped"] += 1

                return
            if self.direction == "max":
                if metric > self.result["best_val"]:
                    self.result["best_val"] = metric
                    self.result["best_id"] = trial_id
                    self.result["best_hp"] = param_string
                if metric < self.result["worst_val"]:
                    self.result["worst_val"] = metric
                    self.result["worst_id"] = trial_id
                    self.result["worst_hp"] = param_string
            elif self.direction == "min":
                if metric < self.result["best_val"]:
                    self.result["best_val"] = metric
                    self.result["best_id"] = trial_id
                    self.result["best_hp"] = param_string
                if metric > self.result["worst_val"]:
                    self.result["worst_val"] = metric
                    self.result["worst_id"] = trial_id
                    self.result["worst_hp"] = param_string

        elif self.experiment_type == "ablation":

            # pop function values and trial_type from parameters, since we don't need them
            param_string.pop("dataset_function", None)
            param_string.pop("model_function", None)
            # First finalized trial
            if self.result.get("best_id", None) is None:
                self.result = {
                    "best_id": trial_id,
                    "best_val": metric,
                    "best_config": param_string,
                    "worst_id": trial_id,
                    "worst_val": metric,
                    "worst_config": param_string,
                    "avg": metric,
                    "metric_list": [metric],
                    "num_trials": 1,
                    "early_stopped": 0,
                }
                return

            # for ablation we always consider 'max' as "the direction"
            if metric > self.result["best_val"]:
                self.result["best_id"] = trial_id
                self.result["best_val"] = metric
                self.result["best_config"] = param_string
            elif metric < self.result["worst_val"]:
                self.result["worst_id"] = trial_id
                self.result["worst_val"] = metric
                self.result["worst_config"] = param_string

        # update results and average regardless of experiment type
        self.result["metric_list"].append(metric)
        self.result["num_trials"] += 1
        self.result["avg"] = sum(self.result["metric_list"]) / float(
            len(self.result["metric_list"])
        )

        if trial.early_stop:
            self.result["early_stopped"] += 1

    def _update_maggy_log(self):
        """Creates the status of a maggy experiment with a progress bar.
        """
        finished = self.result["num_trials"]
        log = ""

        if self.experiment_type == "optimization":
            log = (
                "Maggy Optimization "
                + str(finished)
                + "/"
                + str(self.num_trials)
                + " ("
                + str(self.result["early_stopped"])
                + ") "
                + util._progress_bar(finished, self.num_trials)
                + " - BEST "
                + json.dumps(self.result["best_hp"])
                + " - metric "
                + str(self.result["best_val"])
            )

        elif self.experiment_type == "ablation":
            log = (
                "Maggy Ablation "
                + str(finished)
                + "/"
                + str(self.num_trials)
                + util._progress_bar(finished, self.num_trials)
                + " - BEST Excludes"
                + json.dumps(self.result["best_config"])
                + " - metric "
                + str(self.result["best_val"])
            )

        return log

    def _get_logs(self):
        """Return current experiment status and executor logs to send them to
        spark magic.
        """
        with self.log_lock:
            temp = self.executor_logs
            # clear the executor logs since they are being sent
            self.executor_logs = ""
            return self.result, temp

    def _log(self, log_msg):
        """Logs a string to the maggy driver log file.
        """
        msg = datetime.now().isoformat() + ": " + str(log_msg)
        self.fd.write((msg + "\n").encode())



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import random


class Searchspace(object):
    """Create an instance of `Searchspace` from keyword arguments.

    A searchspace is essentially a set of key value pairs, defining the
    hyperparameters with a name, type and a feasible interval. The keyword
    arguments specify name-values pairs for the hyperparameters,
    where values are tuples of the form (type, list). Type is a string with
    one of the following values:

        - DOUBLE
        - INTEGER
        - DISCRETE
        - CATEGORICAL

    And the list in the tuple specifies either two values only, the start
    and end point of of the feasible interval for DOUBLE and INTEGER,
    or the discrete possible values for the types DISCRETE and CATEGORICAL.

    Sample usage:

    >>> # Define Searchspace
    >>> from maggy import Searchspace
    >>> # The searchspace can be instantiated with parameters
    >>> sp = Searchspace(kernel=('INTEGER', [2, 8]), pool=('INTEGER', [2, 8]))
    >>> # Or additional parameters can be added one by one
    >>> sp.add('dropout', ('DOUBLE', [0.01, 0.99]))

    The `Searchspace` object can also be initialized from a python dictionary:

    >>> sp_dict = sp.to_dict()
    >>> sp_new = Searchspace(**sp_dict)

    The parameter names are added as attributes of `Searchspace` object,
    so they can be accessed directly with the dot notation
    `searchspace._name_`.
    """

    DOUBLE = "DOUBLE"
    INTEGER = "INTEGER"
    DISCRETE = "DISCRETE"
    CATEGORICAL = "CATEGORICAL"

    def __init__(self, **kwargs):
        self._hparam_types = {}
        self._names = []
        for name, value in kwargs.items():
            self.add(name, value)

    def add(self, name, value):
        """Adds {name, value} pair to hyperparameters.

        :param name: Name of the hyperparameter
        :type name: str
        :param value: A tuple of the parameter type and its feasible region
        :type value: tuple
        :raises ValueError: Hyperparameter name is reserved
        :raises ValueError: Hyperparameter feasible region in wrong format
        """
        if getattr(self, name, None) is not None:
            raise ValueError("Hyperparameter name is reserved: {}".format(name))

        if isinstance(value, tuple) or isinstance(value, list):

            if len(value) != 2:
                raise ValueError(
                    "Hyperparameter tuple has to be of length "
                    "two and format (type, list): {0}, {1}".format(name, value)
                )

            param_type = value[0].upper()
            param_values = value[1]

            if param_type in [
                Searchspace.DOUBLE,
                Searchspace.INTEGER,
                Searchspace.DISCRETE,
                Searchspace.CATEGORICAL,
            ]:

                if len(param_values) == 0:
                    raise ValueError(
                        "Hyperparameter feasible region list "
                        "cannot be empty: {0}, {1}".format(name, param_values)
                    )

                if param_type in [Searchspace.DOUBLE, Searchspace.INTEGER]:
                    assert len(param_values) == 2, (
                        "For DOUBLE or INTEGER type parameters, list "
                        "can only contain upper and lower bounds: {0}, {1}".format(
                            name, param_values
                        )
                    )

                    if param_type == Searchspace.DOUBLE:
                        if type(param_values[0]) not in [int, float] or type(
                            param_values[1]
                        ) not in [int, float]:
                            raise ValueError(
                                "Hyperparameter boundaries for type DOUBLE need to be integer "
                                "or float: {}".format(name)
                            )
                    elif param_type == Searchspace.INTEGER:
                        if type(param_values[0]) != int or type(param_values[1]) != int:
                            raise ValueError(
                                "Hyperparameter boundaries for type INTEGER need to be integer: "
                                "{}".format(name)
                            )

                    assert param_values[0] < param_values[1], (
                        "Lower bound {0} must be "
                        "less than upper bound {1}: {2}".format(
                            param_values[0], param_values[1], name
                        )
                    )

                self._hparam_types[name] = param_type
                setattr(self, name, value[1])
                self._names.append(name)
            else:
                raise ValueError(
                    "Hyperparameter type is not of type DOUBLE, "
                    "INTEGER, DISCRETE or CATEGORICAL: {}".format(name)
                )

        else:
            raise ValueError("Value is not an appropriate tuple: {}".format(name))

        print("Hyperparameter added: {}".format(name))

    def to_dict(self):
        """Return the hyperparameters as a Python dictionary.

        :return: A dictionary with hyperparameter names as keys. The values are
            the hyperparameter values.
        :rtype: dict
        """
        return {
            n: (self._hparam_types[n], getattr(self, n))
            for n in self._hparam_types.keys()
        }

    def names(self):
        """Returns the dictionary with the names and types of all
        hyperparameters.

        :return: Dictionary of hyperparameter names, with types as value
        :rtype: dict
        """
        return self._hparam_types

    def get(self, name, default=None):
        """Returns the value of `name` if it exists, else `default`."""
        if name in self._hparam_types:
            return getattr(self, name)

        return default

    def get_random_parameter_values(self, num):
        """Generate random parameter dictionaries, e.g. to be used for initializing an optimizer.

        :param num: number of random parameter dictionaries to be generated.
        :type num: int
        :raises ValueError: `num` is not an int.
        :return: a list containing parameter dictionaries
        :rtype: list
        """
        return_list = []
        for _ in range(num):
            params = {}
            for name, value in self.names().items():
                feasible_region = self.get(name)
                if value == Searchspace.DOUBLE:
                    params[name] = random.uniform(
                        feasible_region[0], feasible_region[1]
                    )
                elif value == Searchspace.INTEGER:
                    params[name] = random.randint(
                        feasible_region[0], feasible_region[1]
                    )
                elif value == Searchspace.DISCRETE:
                    params[name] = random.choice(feasible_region)
                elif value == Searchspace.CATEGORICAL:
                    params[name] = random.choice(feasible_region)
            return_list.append(params)

        return return_list

    def __iter__(self):
        self._returned = self._names.copy()
        return self

    def __next__(self):
        # if list not empty
        if self._returned:
            # pop from left and get parameter tuple
            name = self._returned.pop(0)
            return {
                "name": name,
                "type": self._hparam_types[name],
                "values": self.get(name),
            }
        else:
            raise StopIteration

    def items(self):
        """Returns a sorted iterable over all hyperparameters in the searchspace.

        Allows to iterate over the hyperparameters in a searchspace. The parameters
        are sorted in the order of which they were added to the searchspace by the user.

        :return: an iterable of the searchspace
        :type: Searchspace
        """
        # for consistency and serves mainly as syntactic sugar
        return self

    def keys(self):
        """Returns a sorted iterable list over the names of hyperparameters in
        the searchspace.

        :return: names of hyperparameters as a list of strings
        :type: list
        """
        return self._names

    def values(self):
        """Returns a sorted iterable list over the types and feasible intervals of
        hyperparameters in the searchspace.

        :return: types and feasible interval of hyperparameters as tuple
        :type: tuple
        """
        return [(self._hparam_types[name], self.get(name)) for name in self._names]

    def __contains__(self, name):
        return name in self._hparam_types

    def __str__(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def json(self):
        return json.dumps(self.to_dict(), sort_keys=True)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
Experiment module used for running asynchronous optimization tasks.

The programming model is that you wrap the code containing the model
training inside a wrapper function.
Inside that wrapper function provide all imports and parts that make up your
experiment, see examples below. Whenever a function to run an experiment is
invoked it is also registered in the Experiments service along with the
provided information.
"""
import os
import atexit
import time

from hops import util as hopsutil
from hops.experiment_impl.util import experiment_utils

from maggy import util, tensorboard
from maggy.core import trialexecutor, experimentdriver

app_id = None
running = False
run_id = 1
experiment_json = None


def lagom(
    train_fn,
    name="no-name",
    experiment_type="optimization",
    searchspace=None,
    optimizer=None,
    direction="max",
    num_trials=1,
    ablation_study=None,
    ablator=None,
    optimization_key="metric",
    hb_interval=1,
    es_policy="median",
    es_interval=1,
    es_min=10,
    description="",
):
    """Launches a maggy experiment, which depending on `experiment_type` can
    either be a hyperparameter optimization or an ablation study experiment.
    Given a search space, objective and a model training procedure `train_fn`
    (black-box function), an experiment is the whole process of finding the
    best hyperparameter combination in the search space, optimizing the
    black-box function. Currently maggy supports random search and a median
    stopping rule.

    **lagom** is a Swedish word meaning "just the right amount".

    :param train_fn: User defined experiment containing the model training.
    :type train_fn: function
    :param name: A user defined experiment identifier.
    :type name: str
    :param experiment_type: Type of Maggy experiment, either 'optimization'
        (default) or 'ablation'.
    :type experiment_type: str
    :param searchspace: A maggy Searchspace object from which samples are
        drawn.
    :type searchspace: Searchspace
    :param optimizer: The optimizer is the part generating new trials.
    :type optimizer: str, AbstractOptimizer
    :param direction: If set to ‘max’ the highest value returned will
        correspond to the best solution, if set to ‘min’ the opposite is true.
    :type direction: str
    :param num_trials: the number of trials to evaluate given the search space,
        each containing a different hyperparameter combination
    :type num_trials: int
    :param ablation_study: Ablation study object. Can be None for optimization
        experiment type.
    :type ablation_study: AblationStudy
    :param ablator: Ablator to use for experiment type 'ablation'.
    :type ablator: str, AbstractAblator
    :param optimization_key: Name of the metric to be optimized
    :type optimization_key: str, optional
    :param hb_interval: The heartbeat interval in seconds from trial executor
        to experiment driver, defaults to 1
    :type hb_interval: int, optional
    :param es_policy: The earlystopping policy, defaults to 'median'
    :type es_policy: str, optional
    :param es_interval: Frequency interval in number of steps to check currently
        running trials for early stopping, defaults to 1.
    :type es_interval: int, optional
    :param es_min: Minimum number of trials finalized before checking for
        early stopping, defaults to 10
    :type es_min: int, optional
    :param description: A longer description of the experiment.
    :type description: str, optional
    :raises RuntimeError: An experiment is currently running.
    :return: A dictionary indicating the best trial and best hyperparameter
        combination with it's performance metric
    :rtype: dict
    """
    global running
    if running:
        raise RuntimeError("An experiment is currently running.")

    job_start = time.time()
    sc = hopsutil._find_spark().sparkContext
    exp_driver = None

    try:
        global app_id
        global experiment_json
        global run_id
        app_id = str(sc.applicationId)

        app_id, run_id = util._validate_ml_id(app_id, run_id)

        # start run
        running = True
        experiment_utils._set_ml_id(app_id, run_id)

        # create experiment dir
        experiment_utils._create_experiment_dir(app_id, run_id)

        tensorboard._register(experiment_utils._get_logdir(app_id, run_id))

        num_executors = util.num_executors(sc)

        # start experiment driver
        if experiment_type == "optimization":

            assert num_trials > 0, "number of trials should be greater " + "than zero"
            tensorboard._write_hparams_config(
                experiment_utils._get_logdir(app_id, run_id), searchspace
            )

            if num_executors > num_trials:
                num_executors = num_trials

            exp_driver = experimentdriver.ExperimentDriver(
                "optimization",
                searchspace=searchspace,
                optimizer=optimizer,
                direction=direction,
                num_trials=num_trials,
                name=name,
                num_executors=num_executors,
                hb_interval=hb_interval,
                es_policy=es_policy,
                es_interval=es_interval,
                es_min=es_min,
                description=description,
                log_dir=experiment_utils._get_logdir(app_id, run_id),
            )

            exp_function = exp_driver.optimizer.name()

        elif experiment_type == "ablation":
            exp_driver = experimentdriver.ExperimentDriver(
                "ablation",
                ablation_study=ablation_study,
                ablator=ablator,
                name=name,
                num_executors=num_executors,
                hb_interval=hb_interval,
                description=description,
                log_dir=experiment_utils._get_logdir(app_id, run_id),
            )
            # using exp_driver.num_executor since
            # it has been set using ablator.get_number_of_trials()
            # in experiment.py
            if num_executors > exp_driver.num_executors:
                num_executors = exp_driver.num_executors

            exp_function = exp_driver.ablator.name()
        else:
            running = False
            raise RuntimeError(
                "Unknown experiment_type:"
                "should be either 'optimization' or 'ablation', "
                "But it is '{0}'".format(str(experiment_type))
            )

        nodeRDD = sc.parallelize(range(num_executors), num_executors)

        # Do provenance after initializing exp_driver, because exp_driver does
        # the type checks for optimizer and searchspace
        sc.setJobGroup(os.environ["ML_ID"], "{0} | {1}".format(name, exp_function))

        experiment_json = experiment_utils._populate_experiment(
            name,
            exp_function,
            "MAGGY",
            exp_driver.searchspace.json(),
            description,
            app_id,
            direction,
            optimization_key,
        )

        exp_ml_id = app_id + "_" + str(run_id)
        experiment_json = experiment_utils._attach_experiment_xattr(exp_ml_id, experiment_json, "INIT")

        util._log(
            "Started Maggy Experiment: {0}, {1}, run {2}".format(name, app_id, run_id)
        )

        exp_driver.init(job_start)

        server_addr = exp_driver.server_addr

        # Force execution on executor, since GPU is located on executor
        nodeRDD.foreachPartition(
            trialexecutor._prepare_func(
                app_id,
                run_id,
                experiment_type,
                train_fn,
                server_addr,
                hb_interval,
                exp_driver._secret,
                optimization_key,
                experiment_utils._get_logdir(app_id, run_id),
            )
        )
        job_end = time.time()

        result = exp_driver.finalize(job_end)
        best_logdir = (
            experiment_utils._get_logdir(app_id, run_id) + "/" + result["best_id"]
        )

        util._finalize_experiment(
            experiment_json,
            float(result["best_val"]),
            app_id,
            run_id,
            "FINISHED",
            exp_driver.duration,
            experiment_utils._get_logdir(app_id, run_id),
            best_logdir,
            optimization_key,
        )

        util._log("Finished Experiment")

        return result

    except:  # noqa: E722
        _exception_handler(
            experiment_utils._seconds_to_milliseconds(time.time() - job_start)
        )
        if exp_driver:
            if exp_driver.exception:
                raise exp_driver.exception
        raise
    finally:
        # grace period to send last logs to sparkmagic
        # sparkmagic hb poll intervall is 5 seconds, therefore wait 6 seconds
        time.sleep(6)
        # cleanup spark jobs
        if running and exp_driver is not None:
            exp_driver.stop()
        run_id += 1
        running = False
        sc.setJobGroup("", "")

    return result


def _exception_handler(duration):
    """
    Handles exceptions during execution of an experiment

    :param duration: duration of the experiment until exception in milliseconds
    :type duration: int
    """
    try:
        global running
        global experiment_json
        if running and experiment_json is not None:
            experiment_json["state"] = "FAILED"
            experiment_json["duration"] = duration
            exp_ml_id = app_id + "_" + str(run_id)
            experiment_utils._attach_experiment_xattr(exp_ml_id, experiment_json, "FULL_UPDATE")
    except Exception as err:
        util._log(err)


def _exit_handler():
    """
    Handles jobs killed by the user.
    """
    try:
        global running
        global experiment_json
        if running and experiment_json is not None:
            experiment_json["status"] = "KILLED"
            exp_ml_id = app_id + "_" + str(run_id)
            experiment_utils._attach_experiment_xattr(exp_ml_id, experiment_json, "FULL_UPDATE")
    except Exception as err:
        util._log(err)


atexit.register(_exit_handler)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
import time
import random

from maggy.searchspace import Searchspace
from maggy.optimizer import RandomSearch
from maggy import experiment

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")


def test_randomsearch_init():

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]), param2=("integer", [3, 4]))

    rs = RandomSearch(5, sp, [])

    assert rs.num_trials == 5
    assert rs.searchspace == sp


def test_randomsearch_initialize():

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]), param2=("integer", [3, 4]))

    rs = RandomSearch(5, sp, [])

    rs.initialize()

    assert len(rs.trial_buffer) == 5


def test_rs_initialize2():

    sp = Searchspace(argument_param=("DISCRETE", [1, 5]))

    rs = RandomSearch(5, sp, [])

    with pytest.raises(NotImplementedError) as excinfo:
        rs.initialize()
    assert "Searchspace needs at least one continuous parameter" in str(excinfo.value)


def test_randomsearch(sc):

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]))

    def train(argument_param, reporter):

        print("Entered train function with param {}".format(argument_param))

        for i in range(5):
            acc = i + random.random()
            reporter.broadcast(metric=acc)
            reporter.log("Metric: {}".format(acc))

            time.sleep(argument_param)

        return acc

    result = experiment.lagom(
        train,
        sp,
        optimizer="randomsearch",
        direction="max",
        num_trials=25,
        name="test",
        hb_interval=1,
        es_interval=10,
    )
    assert type(result) == type({})



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
import tensorflow as tf

HOPSWORKS = "HOPSWORKS"
SPARK_ONLY = "SPARK_ONLY"

mode = None
tf_full = tf.__version__.split(".")[0]
# for building the docs since mock object doesn't mock int()
if not isinstance(tf_full, str):
    tf_version = 2
else:
    tf_version = int(tf_full)

try:
    mode = os.environ["HOPSWORKS_VERSION"]
    mode = HOPSWORKS
    print("You are running maggy on Hopsworks.")
except KeyError:
    mode = SPARK_ONLY
    print("You are running maggy in pure Spark mode.")



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
import time
import random

from maggy import Trial


def test_trial_init():

    trial = Trial({"param1": 5, "param2": "ada"})

    exp = {"param1": 5, "param2": "ada"}

    assert trial.params == exp
    assert trial.status == Trial.PENDING
    assert trial.trial_id == "3d1cc9fdb1d4d001"


def test_trial_serialization():

    trial = Trial({"param1": 5, "param2": "ada"})

    exp = {"param1": 5, "param2": "ada"}

    json_str = trial.to_json()

    new_trial = Trial.from_json(json_str)

    assert isinstance(new_trial, Trial)
    assert new_trial.params == exp
    assert new_trial.status == Trial.PENDING
    assert new_trial.trial_id == "3d1cc9fdb1d4d001"



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
from setuptools import setup, find_packages
from maggy.version import __version__


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='maggy',
    version=__version__,
    install_requires=[
        'numpy'
    ],
    extras_require={
        'pydoop': ['pydoop'],
        'tf': ['tensorflow==1.14.0'],
        'docs': [
            'sphinx==1.8.5',
            'sphinx-autobuild',
            'recommonmark',
            'sphinx_rtd_theme',
            'jupyter_sphinx_theme'
        ],
        'test': [
            'pylint',
            'pytest',
        ],
        'dev': [
            'black',
            'flake8',
        ],
        'spark': ['pyspark==2.4.3']
    },
    author='Moritz Meister',
    author_email='meister.mo@gmail.com',
    description='Efficient asynchronous optimization of expensive black-box functions on top of Apache Spark',
    license='GNU Affero General Public License v3',
    keywords='Hyperparameter, Optimization, Auto-ML, Hops, Hadoop, TensorFlow, Spark',
    url='https://github.com/logicalclocks/maggy',
    download_url='http://snurran.sics.se/hops/maggy/maggy-' + __version__ + '.tar.gz',
    packages=find_packages(),
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
    ]
)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
from operator import add

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")

# Can also use a decorator such as this to use specific fixtures in specific functions
# @pytest.mark.usefixtures("spark_context", "hive_context")


def do_word_counts(lines):
    """ count of words in an rdd of lines """

    counts = lines.flatMap(lambda x: x.split()).map(lambda x: (x, 1)).reduceByKey(add)
    results = {word: count for word, count in counts.collect()}
    return results


# start function with test_ so pytest can discover them
def test_do_word_counts(sc):
    """ test that a single event is parsed correctly
    Args:
        spark_context: test fixture SparkContext
        hive_context: test fixture HiveContext
    """

    test_input = [" hello spark ", " hello again spark spark"]

    input_rdd = sc.parallelize(test_input, 1)
    results = do_word_counts(input_rdd)

    expected_results = {"hello": 2, "spark": 3, "again": 1}
    assert results == expected_results



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
from maggy.searchspace import Searchspace
from maggy.optimizer import RandomSearch

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")


def test_nr_executors(sc):

    executor_instances = int(sc._conf.get("spark.executor.instances"))
    expected_number = 2
    assert executor_instances == expected_number


def test_random_search(sc):

    sp = Searchspace(argument_param=("DOUBLE", [1, 5]))

    rs = RandomSearch(5, sp, [])

    exp_result = {"argument_param": "DOUBLE"}

    assert sp.names() == exp_result
    assert rs.num_trials == 5
    assert rs.searchspace == sp
